#ifndef lint
static const char rcsid[] = "$Header: /net/hopkins/home/tristan/mpeg/dec/RCS/simple_player.c,v 1.49 1999/08/03 11:12:07 tristan Exp tristan $";
static const char copyright[] = "Copyright (c) 1997, MpegTV LLC";
#endif


/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee or
 * royalty on an "as is" basis.  Copyright-holder disclaims
 * any and all warranties, whether express, implied, or statuary, including any
 * implied warranties or merchantability or of fitness for a particular
 * purpose.  In no event shall the copyright-holder be liable for any
 * incidental, punitive, or consequential damages of any kind whatsoever
 * arising from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and user's
 * customers, employees, agents, transferees, successors, and assigns.
 *
 */

#include "mtvp/mtvp.h"

#include "forms.h"
#include "simple_player.h"
#include "simple_player_controls.h"

/*--------------------------------------------------------------------------+
| Configuration flags														|
+--------------------------------------------------------------------------*/
/*
 * Define TEST_SIH only to test the sample SIH (Stream Input Handler) plugin.
 * Value of TEST_SIH is path of the sample SIH dynamic library.
 * The sample SIH plugin handles URLs of the form "file:<filename>".
 * Try: $ simple_player file:random.mpg (assuming random.mpg is an MPEG
 * file located in the current directory).
 */
//#define TEST_SIH	"../sih/sample_sih.so"

/*
 * Define TEST_VOH only to test the sample VOH (Video Output Handler) plugin.
 * Value of TEST_VOH is path of the sample VOH dynamic library.
 * The sample VOH plugin uses SDL to play full-screen video.
 */
//#define TEST_VOH	"../voh/sample_voh.so"


/*--------------------------------------------------------------------------+
| Macros																	|
+--------------------------------------------------------------------------*/
/* for the msg.h package */
#define MTVP_MSG_PTR (&Msg)

/*--------------------------------------------------------------------------+
| Data types																|
+--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------+
| Globals																	|
+--------------------------------------------------------------------------*/

/* The form pointers */
FD_controls *Controls;

/*--------------------------------------------------------------------------+
| Statics																	|
+--------------------------------------------------------------------------*/

/* to communicate with player */
static MTVP_PLAYER	*Player;

/* for the msg.h package */
static MTVP_MSG_STRUCT Msg;

/*--------------------------------------------------------------------------+
| my_exit																|
+--------------------------------------------------------------------------*/
static void
my_exit(int n)
{	
	MTVP_MESSAGE(("(pid=%u): exiting, status = %d", (u_int)getpid(), n));
	
	/* terminate (real exit this time) */
	exit(n);
}


/*--------------------------------------------------------------------------+
| at_close_controls															|
+--------------------------------------------------------------------------*/
static int 
at_close_controls(
	FL_FORM *fl_form,
	void	*data
)
{
	/*
	 * This callback is called when the controls form is closed.
	 */
	command(NULL, ACTION_QUIT);
	return FL_OK;
}

/*--------------------------------------------------------------------------+
| quit
+--------------------------------------------------------------------------*/
static int read_command(int fd);

static void 
quit()
{
	/* remove this callback (mtvp_destroy_player will close this pipe) */
	fl_remove_io_callback(mtvp_get_notification_fd(Player), FL_READ, (void (*)(int, void *))read_command);

	/* mtvp_destroy_player() should be called to close pipes to the (now defunct) player */
	mtvp_destroy_player(Player);
	
	my_exit(EXIT_SUCCESS);
}

/*--------------------------------------------------------------------------+
| read_command
+--------------------------------------------------------------------------*/
static int
read_command(int fd)
{
	MTVP_PMP_HEADER	msg;		/* to hold message header from Player */ 
	MTVP_PMP_DATA_A2G	data;	/* to hold message data from Player */
	int		status;
	
	/*
	 * This  callback is called when there is something available for
	 * reading on Player.receive_channel.msg_pipe[MTVP_PMP_FD_READ]
	 * (also equal to fd in this routine).
	 *
	 * This routine should read one message from the Player
	 * and take the appropriate action.
	 *
	 * It is mostly a place holder in this example: This sample
	 * application ignores most messages.  It just checks the
	 * API version number and quits if the message MTVP_NOTIFY_QUIT
	 * is received (which should not happen, by the way!).
	 */
	 
	/* read message from player */
	status = mtvp_read_notification(Player, &msg, &data);
	if (status != 0) {
		/* this can happen if the player is killed brutally */
		MTVP_MESSAGE_ERROR(("a2g_read_msg: error reading msg"));
		quit();
	}

	/* trace message if (Msg_select & MTVP_MSG_INFO) */
	{
		char buf[MTVP_PMP_PRINT_MSG_BUF_SIZE];
		MTVP_MESSAGE(("a2g_read_command: received msg: %s", pmp_print_msg(&msg,  (const char *)&data, buf)));
	}
	
	switch ((MTVP_NOTIFY)msg.type) {
		case MTVP_NOTIFY_STATE:
			/* show Player's state in the form title */
			switch (data.data_generic.data_int) {
			case MTVP_STATE_PAUSED_AT_END:
				fl_set_form_title(Controls->controls, "Simple Player (Paused, end reached)");
				break;
			case MTVP_STATE_PAUSED:
				fl_set_form_title(Controls->controls, "Simple Player (Paused)");
				break;
			case MTVP_STATE_PLAYING:
				fl_set_form_title(Controls->controls, "Simple Player (Playing)");
				break;
			default:
				MTVP_FATAL(("unknow State (should not happen!)"));
			}
			break;
#ifdef TEST_SIH
		case MTVP_NOTIFY_SIH_FEEDBACK:
			{
				u_int payload_size = msg.size - sizeof(int);
				const char *payload_data = (const char *)&data + sizeof(int);
				int	handler_id = data.data_generic.data_int;
				
				printf("received MTVP_NOTIFY_SIH_FEEDBACK from handler_id = %d, size = %u, data = \"%s\"\n", handler_id, payload_size, payload_data);
			}
			break;
#endif /* TEST_SIH */
		case MTVP_NOTIFY_RETURN_FAILURE:
			/*
			 * This message is received each time a command that was sent with
			 * the MTVP_PMP_FLAG_RETURN_STATUS flag failed.
			 *
			 * This is convenient to test the return status of commands like
			 * MTVP_COMMAND_SIH_LINK or MTVP_COMMAND_OPEN_STREAM_URL.
			 *
			 * Note that if MTVP_PMP_FLAG_RETURN_ORIGINAL
			 * was set too, a copy of the original message is sent back right
			 * after the MTVP_NOTIFY_RETURN_FAILURE message.
			 *
			 * Here, in this example, we know that the only command sent with
			 * MTVP_PMP_FLAG_RETURN_STATUS is MTVP_COMMAND_SIH_LINK. Ideally we should check
			 * what was the command associated with this returned status message.
			 */
			printf("command %d failed: error = %d\n", data.data_return_success.original_header.type, data.data_return_failure.error);

			break;
		case MTVP_NOTIFY_RETURN_SUCCESS:
			/*
			 * This message is received each time a command that was sent with
			 * the MTVP_PMP_FLAG_RETURN_STATUS flag succeded.
			 */
			break;
		case MTVP_NOTIFY_QUIT:
			/*
			 * mtvp sends MTVP_NOTIFY_QUIT after receiving MTVP_COMMAND_QUIT and before exiting.
			 * mtv will terminate when it receives the SIGCHLD just after mtvp has exited.
			 */
			quit();

			break;
		default:
			{
				char buf[MTVP_PMP_PRINT_MSG_BUF_SIZE];
				/* msgs not understood should be ignored */
				MTVP_MESSAGE(("a2g_read_command: ignoring msg: %s", pmp_print_msg(&msg,  (const char *)&data, buf)));
			}
			break;
	}

	return 0;
}

/*--------------------------------------------------------------------------+
| open_url																	|
+--------------------------------------------------------------------------*/
void
open_url(const char *url)
{
	int		status;
	void	*value;
	
	mtvp_send_command_string(Player, MTVP_COMMAND_OPEN_STREAM_URL, MTVP_PMP_FLAG_SYNCHRONOUS, url);
	if (mtvp_get_command_status(Player, &status, &value) != 0) {
		abort();	// last command did not have MTVP_PMP_FLAG_SYNCHRONOUS!
	}
	
	if (status != 0) {
		printf("%s: %s\n", url, msg_error_string(status));
	} else {
		printf("%s was opened by SIH header-id = %d\n", url, (int)value);
	}
}

/*--------------------------------------------------------------------------+
| command																	|
+--------------------------------------------------------------------------*/
void
command(
	FL_OBJECT *fl_object,
	long param
)
{
	MTVP_COMMAND	com = param;
	
	/*
	 * This callback is called when clicking on any of the Xforms graphic controls
	 * of the control panel.  The param indicates which control was used.
	 */
	
	MTVP_MESSAGE(("executing command call-back"));
	
	switch (com) {
		case ACTION_QUIT:
		
			/* tell Player to quit */
			mtvp_send_command_void(Player, MTVP_COMMAND_QUIT, 0);

			break;
		case ACTION_PLAY:
		
			/* tell Player to start playing */
			mtvp_send_command_int(Player, MTVP_COMMAND_STATE, 0, MTVP_STATE_PLAYING);
			
			break;
		case ACTION_STEP:
		
			/* tell Player to Step one frame */
			/* Note: if player was playing, state changes automatically to MTVP_STATE_PAUSED */
			mtvp_send_command_int(Player, MTVP_COMMAND_STEPS, 0, 1);
			
			break;
		case ACTION_REWIND:
		
			/* tell Player to Rewind */
			mtvp_send_command_int(Player, MTVP_COMMAND_SEEK, 0, 0);
			
			break;
		case ACTION_OPEN_FILE:
			{
				/* get file name with standard Xforms file selector */
				const char *url = fl_show_fselector ("Open",
										NULL,
										"*.mpg",
										NULL);
				if (url != NULL) {
					open_url(url);
				}
			}
			break;
		case ACTION_OPEN_URL:
			{
				/* ask user to enter URL using Xforms gadget */
				const char *url = fl_show_input ("Enter URL:", "http://");

				if (url != NULL) {
					open_url(url);
				}
			}
			break;
		default:
			/* this hould not happen! */
			MTVP_FATAL(("command callback does not expect command %d", com));
			break;
	}
}

/*--------------------------------------------------------------------------+
| init_form																	|
+--------------------------------------------------------------------------*/
static void
init_form(
	int		argc,
	const char **argv
)
{	
	if (fl_initialize (&argc, (char **)argv, "Simple Player (Stopped)", 0, 0) == NULL) {
		int status = errno;
		MTVP_FATAL_ERROR(("cannot initialize xform"));
	}

	if ((Controls = create_form_controls()) == NULL) {
		int status = errno;
		MTVP_FATAL_ERROR(("cannot create Controls form"));
	}
	
	fl_set_form_atclose(Controls->controls, at_close_controls, NULL);

	fl_show_form(Controls->controls, FL_PLACE_GEOMETRY, FL_FULLBORDER, "Simple Player");

	MTVP_MESSAGE(("adding io call-back"));

	fl_add_io_callback(mtvp_get_notification_fd(Player), FL_READ, (void (*)(int, void *))read_command, NULL);

}



/*--------------------------------------------------------------------------+
| main																		|
+--------------------------------------------------------------------------*/
/*ARGSUSED*/
int
main(
	int argc,
	const char **argv,
	const char **envp
)
{
	const char	**p, **q;
	int		n;
	int		status;

	/*
	 * Set Msg.select == -1 to print all the warnings and messages.
	 * See msg.h for the MTVP_MESSAGE() and WARNING() macros definitions.
	 */
	 
	//Msg.select = -1;

	MTVP_MESSAGE(("started, pid = %u", (u_int)getpid()));
	
	/*
	 * Start the player
	 */
	
	{
		static const char *mtvp_argv[] = {
#ifdef HANDLE_CHALLENGE
			"./mtvp",	/* name of the player (if no / in pathname, will search in PATH) */
#else
			"../../bin/mtvp",	/* name of the player (if no / in pathname, will search in PATH) */
#endif
			//"-w",		/* trace everything */
			//"-a0",	/* disable audio */
			NULL,	/* must be NULL-terminated */
		};

		/*
		 * First parameter of mtvp_create_player() is a ptr to a NULL-terminated array
		 * of strings.
		 * The first string is the name of the Player ("mtvp").
		 * The following strings are possible parameters.
		 * If the first string is a full pathname (starting with "/")
		 * mtvp_create_player() will run this program as "the Player".
		 *
		 * If the first string is not a full pathname (starting with "/")
		 * mtvp_create_player() will look for the Player in the following places:
		 * - in the directory where argv[0] is located (if argv[0] contains a '/')
		 * - in the current directory (".")
		 * - in the PATH
		 *
		 * Typically first string is "mtvp", if the player was installed
		 * under this name.
		 *
		 * Second parameter of mtvp_create_player() is a string that should be
		 * the argv[0] from the calling program.
		 */
	
		if ((status = mtvp_create_player(&Player, mtvp_argv, argv[0], NULL)) != 0) {
			if (status == -1) {
				MTVP_FATAL_ERROR(("cannot start player (%s) - probably mtvp API version mismatch", mtvp_argv[0]));
			} else if (status == EPIPE) {
				MTVP_FATAL_ERROR(("cannot start player (%s) - check that player is in your PATH!", mtvp_argv[0]));
			} else {
				MTVP_FATAL_ERROR(("cannot start player (%s) - could not fork", mtvp_argv[0]));
			}
		}
	}
	
	/* initialize the form */
	init_form(argc, argv);
	
#ifdef TEST_SIH
	{
		int		status;
		void	*value;

		/* link custom stream interface */
		mtvp_send_command_string(Player, MTVP_COMMAND_SIH_LINK, MTVP_PMP_FLAG_SYNCHRONOUS, TEST_SIH);
	
		if (mtvp_get_command_status(Player, &status, &value) != 0) {
			abort_now();	// last command did not have MTVP_PMP_FLAG_SYNCHRONOUS!
		}
		
		if (status != 0) {
			printf("cannot link SIH plugin \"%s\"\n", TEST_SIH);	
		} else {		
			int	handler_id = (int)value;
			printf("SIH plugin \"%s\" linked, handler_id = %d\n", TEST_SIH, handler_id);
			{
				struct {
					int	sih_handler_id;
					char buf[256];
				} sih_cntl_data;
				
				sih_cntl_data.sih_handler_id = handler_id;
				strcpy(sih_cntl_data.buf, "testing MTVP_COMMAND_SIH_CNTL");
				mtvp_send_command_custom(Player, MTVP_COMMAND_SIH_CNTL, 0, (const void *)&sih_cntl_data, sizeof(int) + strlen(sih_cntl_data.buf) + 1);
			}
		}
	}

#endif /* TEST_SIH */
	
#ifdef TEST_VOH
	{
		int		status;
		void	*value;
	
		/* link custom stream interface */
		mtvp_send_command_string(Player, MTVP_COMMAND_VOH_LINK, MTVP_PMP_FLAG_SYNCHRONOUS, TEST_VOH);

		if (mtvp_get_command_status(Player, &status, &value) != 0) {
			abort_now();	// last command did not have MTVP_PMP_FLAG_SYNCHRONOUS!
		}

		if (status != 0) {
			MTVP_FATAL_ERROR(("cannot link VOH plugin"));
		} else {		
			int	handler_id = (int)value;
			printf("VOH plugin linked successfully, handler_id = %d\n", handler_id);
			mtvp_send_command_int(Player, MTVP_COMMAND_VOH_SELECT, MTVP_PMP_FLAG_SYNCHRONOUS, handler_id);

			if (mtvp_get_command_status(Player, &status, &value) != 0) {
				abort_now();	// last command did not have MTVP_PMP_FLAG_SYNCHRONOUS!
			}

			if (status != 0) {
				MTVP_FATAL_ERROR(("cannot select VOH plugin"));
			} else {
				printf("VOH plugin selected\n");
			}
		
			{
				struct {
					int	voh_handler_id;
					char buf[256];
				} voh_cntl_data;
				
				voh_cntl_data.voh_handler_id = handler_id;
				strcpy(voh_cntl_data.buf, "testing MTVP_COMMAND_VOH_CNTL");
				mtvp_send_command_custom(Player, MTVP_COMMAND_VOH_CNTL, 0, (const void *)&voh_cntl_data, sizeof(int) + strlen(voh_cntl_data.buf) + 1);
			}
		}
	}
#endif /* TEST_VOH */
	
	/* if one argument was passed, assume that it is a URL to open */
	if (argc == 2) {

		/* open stream from URL */
		open_url(argv[1]);
		
	} else if (argc > 2) {
		fprintf(stderr, "usage: simple_player [URL]\n");
		fprintf(stderr, "example: simple_player file:movie.mpg\n");
		my_exit(EXIT_FAILURE);
	}

	MTVP_MESSAGE(("entering event loop"));

	/* enter event loop */
	fl_do_forms();
	
	MTVP_MESSAGE(("What?!? fl_do_forms() returned! (should never happen)"));

	/* terminated successfully */
	my_exit(EXIT_SUCCESS);

	/* so that lint does not complain! */
	return 0;
}


/* eof */
